from pydantic import BaseModel, Field

from src.domain._shared import optional_field
from src.domain.todo import TodoEntity


class UpdateTodoBodyDTO(BaseModel):
    title: str | None = optional_field(TodoEntity, "title")
    description: str | None = optional_field(TodoEntity, "description")
    done: bool | None = optional_field(TodoEntity, "done")


class UpdateTodoParamDTO(BaseModel):
    id: str = Field(description="The ID of the todo to update")



from abc import ABC, abstractmethod


class Usecase[TResult, TInput = None](ABC):
    @abstractmethod
    async def execute(self, *, data: TInput = None) -> TResult:
        raise NotImplementedError



from pydantic import BaseModel, Field

from dijay import injectable
from src.application._shared import Usecase
from src.domain.todo import TodoEntity, TodoRepository


class ListTodoResult(BaseModel):
    items: list[TodoEntity] = Field(
        default_factory=list,
        description="List of todos",
    )


@injectable()
class ListTodoUsecase(Usecase[ListTodoResult]):
    def __init__(self, todo_repository: TodoRepository):
        self.todo_repository = todo_repository

    async def execute(self) -> ListTodoResult:
        items = await self.todo_repository.find_all()
        return ListTodoResult(
            items=items,
        )



"""Global container instance and module-level shortcuts."""

from .container import Container

_global = Container()

injectable = _global.injectable
"""Shortcut for :meth:`Container.injectable` on the global container."""

register = _global.register
"""Shortcut for :meth:`Container.register` on the global container."""

resolve = _global.resolve
"""Shortcut for :meth:`Container.resolve` on the global container."""

on_bootstrap = _global.on_bootstrap
"""Shortcut for :meth:`Container.on_bootstrap` on the global container."""

on_shutdown = _global.on_shutdown
"""Shortcut for :meth:`Container.on_shutdown` on the global container."""


def instance() -> Container:
    """Create and return a new, independent :class:`Container`.

    Use this when you need isolation from the global container,
    for example in tests or multi-tenant scenarios.
    """
    return Container()



from pydantic import BaseModel, Field

from dijay import injectable
from src.application._shared import Usecase
from src.domain._shared import optional_field
from src.domain.todo import TodoEntity, TodoNotFoundError, TodoRepository


class UpdateTodoData(BaseModel):
    class Changes(BaseModel):
        title: str | None = optional_field(TodoEntity, "title")
        description: str | None = optional_field(TodoEntity, "description")
        done: bool | None = optional_field(TodoEntity, "done")

    id: str = TodoEntity.model_fields["id"]
    changes: Changes = Field(
        description="Changes to be applied to the todo",
        examples=[
            Changes(
                title="Buy milk",
                description="Buy milk",
                done=True,
            )
        ],
    )


@injectable()
class UpdateTodoUsecase(Usecase[TodoEntity, UpdateTodoData]):
    def __init__(self, todo_repository: TodoRepository):
        self.todo_repository = todo_repository

    async def execute(self, data: UpdateTodoData) -> TodoEntity:
        todo = await self.todo_repository.find_by_id(data.id)
        if not todo:
            raise TodoNotFoundError()

        update_data = data.changes.model_dump(exclude_unset=True)
        updated_todo = todo.model_copy(update=update_data)

        await self.todo_repository.save(updated_todo)
        return updated_todo



import asyncio
from typing import Annotated, Any

import pytest

from dijay import Container, Inject, injectable, instance, on_bootstrap, on_shutdown


@pytest.mark.asyncio
async def test_method_lifecycle_hooks():
    c = instance()
    events = []

    @c.injectable()
    class Service:
        @on_bootstrap
        async def boot(self):
            events.append("boot_started")
            await asyncio.sleep(0.01)
            events.append("boot_done")

        @on_shutdown
        def shut(self):
            events.append("shut")

    async with c:
        assert "boot_done" in events
        assert "shut" not in events

    assert "shut" in events


@pytest.mark.asyncio
async def test_method_hooks_with_injection():
    c = instance()
    events = []

    @c.injectable()
    class Dep:
        pass

    @c.injectable()
    class Service:
        @on_bootstrap
        async def boot(self, dep: Dep):
            assert isinstance(dep, Dep)
            events.append("boot")

    async with c:
        pass

    assert "boot" in events


@pytest.mark.asyncio
async def test_standalone_hooks_still_work():
    c = instance()
    events = []

    @c.on_bootstrap
    def boot():
        events.append("boot")

    async with c:
        pass

    assert "boot" in events


@pytest.mark.asyncio
async def test_register_with_lifecycle_hooks():
    c = instance()
    events = []

    class ManualService:
        @on_bootstrap
        def init(self):
            events.append("init")

        @on_shutdown
        async def close(self):
            events.append("close")

    c.register(ManualService, ManualService)

    async with c:
        assert "init" in events

    assert "close" in events


@pytest.mark.asyncio
async def test_shutdown_skips_uncreated_singletons():
    c = instance()
    events = []

    @c.injectable()
    class UnusedService:
        @on_shutdown
        def shut(self):
            events.append("shut")

    # We start and stop without ever resolving UnusedService
    async with c:
        pass

    assert "shut" not in events


@pytest.mark.asyncio
async def test_multiple_lifecycle_methods():
    c = instance()
    events = []

    @c.injectable()
    class MultiService:
        @on_bootstrap
        def boot1(self):
            events.append("boot1")

        @on_bootstrap
        def boot2(self):
            events.append("boot2")

    async with c:
        assert "boot1" in events
        assert "boot2" in events


class GlobalService:
    def method(self):
        pass


@pytest.mark.asyncio
async def test_hook_skip_logic():
    c = instance()

    # Manually add a "method" to hooks to trigger the 'continue'
    # Use a class defined at module level to avoid '<locals>'
    c._bootstrap_hooks.append(GlobalService.method)
    c._shutdown_hooks.append(GlobalService.method)

    async with c:
        pass  # Should skip GlobalService.method and not crash


@pytest.mark.asyncio
async def test_async_shutdown_hook():
    c = instance()
    events = []

    @c.on_shutdown
    async def shut():
        events.append("shut")

    async with c:
        pass

    assert "shut" in events


@pytest.mark.asyncio
async def test_unregistered_class_auto_resolve_in_lifecycle():
    c = instance()

    class Standalone:
        pass

    obj = await c.resolve(Standalone)
    assert isinstance(obj, Standalone)


@pytest.mark.asyncio
async def test_any_and_optional_in_injectable():
    c = instance()

    @c.injectable()
    class Service:
        def __init__(self, x: Any = 1, y: str | None = None):
            self.x = x
            self.y = y

    s = await c.resolve(Service)
    assert s.x == 1
    assert s.y is None


@pytest.mark.asyncio
async def test_annotated_without_inject_in_lifecycle():
    c = instance()

    @c.injectable()
    class Service:
        def __init__(self, x: Annotated[int, "metadata"] = 10):
            self.x = x

    s = await c.resolve(Service)
    # It tries to resolve int, which is a class, returns 0
    assert s.x == 0


@pytest.mark.asyncio
async def test_circular_dependency_in_hook():
    c = instance()

    @c.injectable()
    class A:
        def __init__(self, b: B):
            pass

    @c.injectable()
    class B:
        def __init__(self, a: A):
            pass

    @c.on_bootstrap
    async def boot(a: A):
        pass

    with pytest.raises(RuntimeError, match="Circular dependency"):
        async with c:
            pass


@pytest.mark.asyncio
async def test_unregistered_token_in_hook():
    c = instance()

    @c.on_bootstrap
    async def boot(missing: Annotated[str, Inject("unknown")]):
        pass

    with pytest.raises(RuntimeError, match="não registrado"):
        async with c:
            pass


@pytest.mark.asyncio
async def test_lifecycle_with_modules():
    from dijay import module

    events = []

    @injectable()
    class Service:
        @on_bootstrap
        def boot(self):
            events.append("boot")

    @module(providers=[Service])
    class AppModule:
        pass

    c = Container.from_module(AppModule)
    async with c:
        assert "boot" in events



from typing import Annotated

from fastapi import Depends, Request


def inject[T](token: type[T]) -> T:
    """
    Dependency injection helper for FastAPI that resolves a token from the application
    container.

    Args:
        token: The type or class to be resolved.

    Returns:
        The resolved instance of the specified type.
    """

    async def use(request: Request) -> T:
        return await request.app.state.container.resolve(token, id=str(id(request)))

    return Annotated[token, Depends(use)]



import inspect
from collections.abc import Callable
from typing import Any, TypedDict

from .container import Container
from .provider import SINGLETON, Provide


class DynamicModule(TypedDict, total=False):
    """Structure returned by static methods (e.g. ``for_root``) to
    configure dynamic modules.
    """

    module: Any
    providers: list[Any]
    imports: list[Any]
    exports: list[Any]
    globals: bool


def module(
    providers: list[Any] = [],
    imports: list[Any] = [],
    exports: list[Any] = [],
    globals: bool = False,
) -> Callable[[Any], Any]:
    """Decorator that marks a class as a Module.

    Modules organize providers and can be imported by other modules.

    Args:
        providers: Providers to register within this module.
        imports: Imported modules that export providers.
        exports: Providers to export to importing modules.
        globals: If ``True``, exports are available globally.
    """

    def decorator(cls: Any) -> Any:
        cls.__module_metadata__ = {
            "providers": providers,
            "imports": imports,
            "exports": exports,
            "globals": globals,
        }
        return cls

    return decorator


def _is_dynamic_module(obj: Any) -> bool:
    return isinstance(obj, dict) and "module" in obj


def _register_provider(container: Container, provider: Any) -> None:
    """Register a single provider entry into the container.

    Handles three cases:

    - ``Provide`` dataclass with ``use_value``, ``use_class``
      or ``use_factory``.
    - Class decorated with ``@injectable(token)`` carrying
      ``__dijay_token__`` metadata.
    - Plain class (registered as its own token).
    """
    if isinstance(provider, Provide):
        if provider.use_value is not None:
            val = provider.use_value
            container.register(
                provider.token,
                lambda v=val: v,
                scope=provider.scope,
            )
        elif provider.use_factory is not None:
            container.register(
                provider.token,
                provider.use_factory,
                scope=provider.scope,
            )
        elif provider.use_class is not None:
            container.register(
                provider.token,
                provider.use_class,
                scope=provider.scope,
            )
        return

    token = getattr(provider, "__dijay_token__", provider)
    scope = getattr(provider, "__dijay_scope__", SINGLETON)

    if inspect.isclass(provider):
        container.register(token, provider, scope=scope)


def _resolve_module(container: Container, mod: Any) -> None:
    """Recursively resolve a module hierarchy and register
    all providers.
    """
    if _is_dynamic_module(mod):
        for imp in mod.get("imports", []):
            _resolve_module(container, imp)
        for provider in mod.get("providers", []):
            _register_provider(container, provider)
        mod = mod["module"]

    if not hasattr(mod, "__module_metadata__"):
        return

    metadata = mod.__module_metadata__

    for imp in metadata["imports"]:
        _resolve_module(container, imp)

    for provider in metadata["providers"]:
        _register_provider(container, provider)



import os
from collections.abc import Callable

from pydantic import BaseModel, Field

from dijay import injectable


def _get_env_bool(key: str, default: bool) -> Callable[[], bool]:
    def wrapper() -> bool:
        return bool(os.getenv(key, str(default)).lower() == "true")

    return wrapper


@injectable()
class HttpConfig(BaseModel):
    path: str = Field(
        default_factory=lambda: os.getenv("HTTP_PATH", "/"),
        description="Path to the HTTP server",
    )
    host: str = Field(
        default_factory=lambda: os.getenv("HTTP_HOST", "127.0.0.1"),
        description="Host to the HTTP server",
    )
    port: int = Field(
        default_factory=lambda: int(os.getenv("HTTP_PORT", 8000)),
        description="Port to the HTTP server",
    )
    enable_reload: bool = Field(
        default_factory=_get_env_bool("HTTP_ENABLE_RELOAD", True),
        description="Enable reload",
    )
    enable_factory: bool = Field(
        default_factory=_get_env_bool("HTTP_ENABLE_FACTORY", False),
        description="Enable factory",
    )
    title: str = Field(
        default_factory=lambda: os.getenv("HTTP_TITLE", "Todo API"),
        description="Title",
    )
    version: str = Field(
        default_factory=lambda: os.getenv("HTTP_VERSION", "0.0.1"),
        description="Version",
    )
    description: str = Field(
        default_factory=lambda: os.getenv("HTTP_DESCRIPTION", "Todo API"),
        description="Description",
    )



import asyncio

import uvicorn
from watchfiles import run_process

from dijay import Container

from .module import AppModule
from .presentation.http.config import HttpConfig
from .presentation.http.server import HttpServer


async def main():
    async with Container.from_module(AppModule) as container:
        server = await container.resolve(HttpServer)
        config = uvicorn.Config(
            server.app,
            host=server.config.host,
            port=server.config.port,
        )
        await uvicorn.Server(config).serve()


config = HttpConfig()

if config.enable_reload:
    run_process("./src", target=lambda: asyncio.run(main()))
else:
    asyncio.run(main())
